# Copyright 2020 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

from ..exceptions import TsValueError
from ..constants import MAX_NANOSEC


def _parse_seconds_fraction(frac: str) -> int:
    """ Parse the fraction part of a timestamp seconds, using maximum 9 digits
    Returns the nanoseconds
    """
    ns = 0
    mult = MAX_NANOSEC
    for c in frac:
        if c < '0' or c > '9' or int(mult) < 1:
            break
        mult = mult // 10
        ns += mult * int(c)
    return ns


def _parse_iso8601(iso8601: str) -> Tuple[int, int, int, int, int, int, int]:
    """ Limited ISO 8601 timestamp parse; expands YYYY-MM-DDThh:mm:ss.s
    Returns tuple of (year, month, day, hours, mins, seconds, nanoseconds)
    """
    iso_date_time = iso8601.split("T")
    if len(iso_date_time) != 2:
        raise TsValueError("invalid or unsupported ISO 8601 UTC format")
    iso_date = iso_date_time[0].split("-")
    iso_time = iso_date_time[1].split(":")
    if len(iso_date) != 3 or len(iso_time) != 3:
        raise TsValueError("invalid or unsupported ISO 8601 UTC format")
    sec_frac = iso_time[2].split(".")
    if len(sec_frac) != 1 and len(sec_frac) != 2:
        raise TsValueError("invalid or unsupported ISO 8601 UTC format")
    sec = sec_frac[0]
    ns = 0
    if len(sec_frac) > 1:
        ns = _parse_seconds_fraction(sec_frac[1])
    return (int(iso_date[0]), int(iso_date[1]), int(iso_date[2]), int(iso_time[0]), int(iso_time[1]), int(sec), ns)



# Copyright 2019 British Broadcasting Corporation
#
# This is an internal BBC tool and is not licensed externally
# If you have received a copy of this erroneously then you do
# not have permission to reproduce it.

import unittest
from fractions import Fraction

from mediatimestamp import (
    TimeOffset,
    Timestamp,
    SupportsMediaTimeOffset,
    mediatimeoffset,
    SupportsMediaTimestamp,
    mediatimestamp,
    TimeValue)


class TestTimeValue(unittest.TestCase):
    def test_from_int(self):
        tv = TimeValue(100, rate=Fraction(25))
        self.assertEqual(tv.value, 100)
        self.assertEqual(tv.rate, Fraction(25))

    def test_from_timeoffset(self):
        tv = TimeValue(TimeOffset(4), rate=None)
        self.assertEqual(tv.value, TimeOffset(4))
        self.assertIsNone(tv.rate)

    def test_from_timestamp(self):
        tv = TimeValue(Timestamp(4), rate=None)
        self.assertEqual(tv.value, Timestamp(4))
        self.assertIsNone(tv.rate)

    def test_from_timevalue(self):
        tv_in = TimeValue(100, rate=Fraction(25))
        tv = TimeValue(tv_in)
        self.assertEqual(tv.value, 100)
        self.assertEqual(tv.rate, Fraction(25))

    def test_from_timeoffset_to_count(self):
        tv = TimeValue(TimeOffset(4), rate=Fraction(25))
        self.assertIsInstance(tv.value, int)
        self.assertEqual(tv.value, 100)
        self.assertEqual(tv.rate, Fraction(25))

    def test_from_timevalue_rate_change(self):
        tv_in = TimeValue(100, rate=Fraction(25))
        tv = TimeValue(tv_in, rate=Fraction(100))
        self.assertEqual(tv.value, 400)
        self.assertEqual(tv.rate, Fraction(100))

    def test_unsupported_type(self):
        with self.assertRaises(TypeError):
            TimeValue(str(10))

    def test_as_timeoffset(self):
        tv = TimeValue(TimeOffset(4), rate=Fraction(25))
        to = tv.as_timeoffset()
        self.assertIsInstance(to, TimeOffset)
        self.assertEqual(to, TimeOffset(4))

        tv = TimeValue(Timestamp(4), rate=Fraction(25))
        to = tv.as_timeoffset()
        self.assertIsInstance(to, TimeOffset)
        self.assertEqual(to, TimeOffset(4))

        tv = TimeValue(100, rate=Fraction(25))
        to = tv.as_timeoffset()
        self.assertIsInstance(to, TimeOffset)
        self.assertEqual(to, TimeOffset(4))

    def test_mediatimeoffset(self):
        tv = TimeValue(TimeOffset(4), rate=Fraction(25))
        self.assertIsInstance(tv, SupportsMediaTimeOffset)
        to = mediatimeoffset(tv)
        self.assertIsInstance(to, TimeOffset)
        self.assertEqual(to, TimeOffset(4))

        tv = TimeValue(Timestamp(4), rate=Fraction(25))
        self.assertIsInstance(tv, SupportsMediaTimeOffset)
        to = mediatimeoffset(tv)
        self.assertIsInstance(to, TimeOffset)
        self.assertEqual(to, TimeOffset(4))

        tv = TimeValue(100, rate=Fraction(25))
        self.assertIsInstance(tv, SupportsMediaTimeOffset)
        to = mediatimeoffset(tv)
        self.assertIsInstance(to, TimeOffset)
        self.assertEqual(to, TimeOffset(4))

    def test_as_timestamp(self):
        tv = TimeValue(Timestamp(4), rate=Fraction(25))
        ts = tv.as_timestamp()
        self.assertIsInstance(ts, Timestamp)
        self.assertEqual(ts, Timestamp(4))

        tv = TimeValue(TimeOffset(4), rate=Fraction(25))
        ts = tv.as_timestamp()
        self.assertIsInstance(ts, Timestamp)
        self.assertEqual(ts, Timestamp(4))

        tv = TimeValue(100, rate=Fraction(25))
        ts = tv.as_timestamp()
        self.assertIsInstance(ts, Timestamp)
        self.assertEqual(ts, Timestamp(4))

    def test_mediatimestamp(self):
        tv = TimeValue(Timestamp(4), rate=Fraction(25))
        self.assertIsInstance(tv, SupportsMediaTimestamp)
        ts = mediatimestamp(tv)
        self.assertIsInstance(ts, Timestamp)
        self.assertEqual(ts, Timestamp(4))

        tv = TimeValue(TimeOffset(4), rate=Fraction(25))
        self.assertIsInstance(tv, SupportsMediaTimestamp)
        ts = mediatimestamp(tv)
        self.assertIsInstance(ts, Timestamp)
        self.assertEqual(ts, Timestamp(4))

        tv = TimeValue(100, rate=Fraction(25))
        self.assertIsInstance(tv, SupportsMediaTimestamp)
        ts = mediatimestamp(tv)
        self.assertIsInstance(ts, Timestamp)
        self.assertEqual(ts, Timestamp(4))

    def test_as_count(self):
        tv = TimeValue(100)
        ct = tv.as_count()
        self.assertEqual(ct, 100)

        tv = TimeValue(TimeOffset(4), rate=Fraction(25))
        ct = tv.as_count()
        self.assertEqual(ct, 100)

    def test_as_but_no_rate(self):
        tv = TimeValue(TimeOffset(4))
        with self.assertRaises(ValueError):
            tv.as_count()

        tv = TimeValue(100)
        with self.assertRaises(ValueError):
            tv.as_timeoffset()

    def test_from_str(self):
        cases = [
            ("-100", TimeValue(-100)),
            ("0", TimeValue(0)),
            ("100", TimeValue(100)),

            ("100@25", TimeValue(100, rate=Fraction(25))),
            ("100@30000/1001", TimeValue(100, rate=Fraction(30000, 1001))),

            ("-4:0", TimeValue(TimeOffset(4, sign=-1))),
            ("0:0", TimeValue(TimeOffset(0))),
            ("4:0", TimeValue(TimeOffset(4))),

            ("4:0@25", TimeValue(100, rate=Fraction(25))),
            ("4:0@30000/1001", TimeValue(120, rate=Fraction(30000, 1001))),
        ]

        for case in cases:
            with self.subTest(case=case):
                self.assertEqual(TimeValue.from_str(case[0]), case[1])

    def test_from_str_rate(self):
        tv = TimeValue.from_str("100@25", rate=Fraction(100))
        self.assertEqual(tv, TimeValue(100, rate=Fraction(25)))

        tv = TimeValue.from_str("100", rate=Fraction(100))
        self.assertEqual(tv, TimeValue(100, rate=Fraction(100)))

    def test_from_str_invalid(self):
        cases = [
            "100@25@",
            "100/30000/1001",
            "abc",
        ]

        for case in cases:
            with self.subTest(case=case):
                with self.assertRaises(ValueError):
                    TimeValue.from_str(case)

    def test_to_str(self):
        cases = [
            ("-100", TimeValue(-100), True),
            ("0", TimeValue(0), True),
            ("100", TimeValue(100), True),

            ("100@25", TimeValue(100, rate=Fraction(25)), True),
            ("100", TimeValue(100, rate=Fraction(25)), False),
            ("100@30000/1001", TimeValue(100, rate=Fraction(30000, 1001)), True),

            ("-4:0", TimeValue(TimeOffset(4, sign=-1)), True),
            ("0:0", TimeValue(TimeOffset(0)), True),
            ("4:0", TimeValue(TimeOffset(4)), True),
        ]

        for case in cases:
            with self.subTest(case=case):
                self.assertEqual(case[0], case[1].to_str(include_rate=case[2]))

    def test_compare(self):
        self.assertEqual(TimeValue(1), TimeValue(1))
        self.assertNotEqual(TimeValue(1), TimeValue(2))
        self.assertLess(TimeValue(1), TimeValue(2))
        self.assertLessEqual(TimeValue(1), TimeValue(1))
        self.assertGreater(TimeValue(2), TimeValue(1))
        self.assertGreaterEqual(TimeValue(2), TimeValue(2))
        self.assertNotEqual(TimeValue(2), TimeValue(3))
        self.assertEqual(TimeValue(TimeOffset(4)), TimeValue(TimeOffset(4)))

    def test_compare_with_convert(self):
        self.assertEqual(TimeValue(100, rate=Fraction(25)), TimeValue(TimeOffset(4)))
        self.assertEqual(TimeValue(TimeOffset(4)), TimeValue(100, rate=Fraction(25)))

    def test_compare_no_rate(self):
        with self.assertRaises(ValueError):
            TimeValue(100) == TimeValue(TimeOffset(4))

    def test_equality_none(self):
        none_value = None
        self.assertFalse(TimeValue(1) == none_value)
        self.assertTrue(TimeValue(1) != none_value)

    def test_addsub(self):
        cases = [
            (TimeValue(50), '+', TimeValue(50),
                TimeValue(100)),
            (TimeValue(50, rate=Fraction(25)), '+', TimeValue(TimeOffset(2)),
                TimeValue(100, rate=Fraction(25))),
            (TimeValue(TimeOffset(2)), '+', TimeValue(TimeOffset(2)),
                TimeValue(TimeOffset(4))),

            (TimeValue(50), '-', TimeValue(50),
                TimeValue(0)),
            (TimeValue(50, rate=Fraction(25)), '-', TimeValue(TimeOffset(2)),
                TimeValue(0, rate=Fraction(25))),
            (TimeValue(TimeOffset(2)), '-', TimeValue(TimeOffset(2)),
                TimeValue(TimeOffset(0))),
        ]

        for case in cases:
            with self.subTest(case=case):
                if case[1] == '+':
                    result = case[0] + case[2]
                else:
                    result = case[0] - case[2]

                self.assertEqual(result, case[3],
                                 msg="{!r} {} {!r} = {!r}, expected {!r}".format(
                                     case[0], case[1], case[2], result, case[3]))

    def test_addsub_no_rate(self):
        cases = [
            (TimeValue(50), '+', TimeValue(TimeOffset(2))),
            (TimeValue(TimeOffset(2)), '+', TimeValue(50)),

            (TimeValue(50), '-', TimeValue(TimeOffset(2))),
            (TimeValue(TimeOffset(2)), '-', TimeValue(50)),
        ]

        for case in cases:
            with self.subTest(case=case):
                with self.assertRaises(ValueError):
                    if case[1] == '+':
                        case[0] + case[2]
                    else:
                        case[0] - case[2]

    def test_multdiv(self):
        cases = [
            (TimeValue(50), '*', 2,
                TimeValue(100)),
            (TimeValue(TimeOffset(2)), '*', 2,
                TimeValue(TimeOffset(4))),

            (2, '*', TimeValue(50),
                TimeValue(100)),
            (2, '*', TimeValue(TimeOffset(2)),
                TimeValue(TimeOffset(4))),

            (TimeValue(50), '/', 2,
                TimeValue(25)),
            (TimeValue(TimeOffset(2)), '/', 2,
                TimeValue(TimeOffset(1))),

            (TimeValue(25), '/', 2,
                TimeValue(12)),
            (TimeValue(25), '//', 2,
                TimeValue(12)),
        ]

        for case in cases:
            with self.subTest(case=case):
                if case[1] == '*':
                    result = case[0] * case[2]
                elif case[1] == '/':
                    result = case[0] / case[2]
                else:
                    result = case[0] // case[2]

                self.assertEqual(result, case[3],
                                 msg="{!r} {} {!r} = {!r}, expected {!r}".format(
                                     case[0], case[1], case[2], result, case[3]))

    def test_multdiv_not_int(self):
        cases = [
            (TimeValue(50), '*', TimeValue(50)),
            (TimeValue(50), '/', TimeValue(50)),
            (TimeValue(50), '//', TimeValue(50)),
        ]

        for case in cases:
            with self.subTest(case=case):
                with self.assertRaises(TypeError):
                    if case[1] == '*':
                        case[0] * case[2]
                    elif case[1] == '/':
                        case[0] / case[2]
                    else:
                        case[0] // case[2]

    def test_immutable(self):
        tv = TimeValue(0)
        with self.assertRaises(ValueError):
            tv._value = 1

        with self.assertRaises(ValueError):
            tv._rate = Fraction(50)

    def test_hashable(self):
        tv1 = TimeValue(0)
        tv2 = TimeValue.from_str("0:20000000000@50")
        self.assertNotEqual(hash(tv1), hash(tv2))



# Copyright 2020 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This library provides a class TimeOffset which stores an immutable signed time difference value with nanosecond
precision.

It also provides a class Timestamp which is a descendent of TimeOffset which represents an immutable time offset since
the epoch (ie. 1970-01-01T00:00:00.000000000Z)

And finally it includes an immutable TimeRange object which stores a range between two Timestamps.

These data types are of use in a number of situations, but particularly for code that will handle PTP timestamps, which
are normally stored in this fashion.



Expected Logic for binary operations on timestamps and time offsets:

Timestamp and TimeOffset objects can be added and subtracted. The type of the final result depends upon the order and
type of the operands

TS + TO = TS
TS + TS = TS (treats 2nd TS as TO)
TO + TS = TS
TO + TO = TO

TS - TO = TS
TS - TS = TO
TO - TS = TO (treats TS as TO)
TO - TO = TO

+= and -= always give the same result type as the first operand

Instances of TimeOffset can always be multiplied or divided by integers and floats and always give
another TimeOffset as a result.

An instance of Timestamp multiplied by integers or floats will be treated as an instance of TimeOffset
"""

from .timeoffset import TimeOffset, SupportsMediaTimeOffset, mediatimeoffset
from .timestamp import Timestamp, SupportsMediaTimestamp, mediatimestamp
from .timerange import TimeRange, SupportsMediaTimeRange, mediatimerange
from ..exceptions import TsValueError


__all__ = [
    "TimeOffset", "SupportsMediaTimeOffset", "mediatimeoffset",
    "Timestamp", "SupportsMediaTimestamp", "mediatimestamp",
    "TimeRange", "SupportsMediaTimeRange", "mediatimerange",
    "TsValueError"]



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import calendar
import time
import re
from datetime import datetime
from dateutil import tz
from fractions import Fraction
from typing import Tuple, Optional, cast, TYPE_CHECKING
from typing_extensions import Protocol, runtime_checkable

from ..constants import UTC_LEAP
from ..exceptions import TsValueError

from ._parse import _parse_iso8601
from ._types import RationalTypes
from .timeoffset import TimeOffset, TimeOffsetConstructionType, mediatimeoffset

if TYPE_CHECKING:
    from .timerange import TimeRange  # noqa: F401

__all__ = ["Timestamp", "SupportsMediaTimestamp", "mediatimestamp"]


@runtime_checkable
class SupportsMediaTimestamp (Protocol):
    def __mediatimestamp__(self) -> "Timestamp":
        ...


def mediatimestamp(v: SupportsMediaTimestamp) -> "Timestamp":
    """This method can be called on any object which supports the __mediatimestamp__ magic method
    and also on a Timestamp. It will always return a Timestamp or raise a ValueError.
    """
    if isinstance(v, Timestamp):
        return v
    elif hasattr(v, "__mediatimestamp__"):
        return v.__mediatimestamp__()
    else:
        raise ValueError("{!r} cannot be converted to a mediatimestamp.Timestamp".format(v))


class Timestamp(TimeOffset):
    """A nanosecond precision immutable timestamp."""
    def __init__(self, sec: int = 0, ns: int = 0, sign: int = 1):
        super(Timestamp, self).__init__(sec, ns, sign)

    def __setattr__(self, name: str, value: object) -> None:
        raise TsValueError("Cannot assign to an immutable Timestamp")

    def __mediatimeoffset__(self) -> TimeOffset:
        return self

    def __mediatimestamp__(self) -> "Timestamp":
        return self

    def __mediatimerange__(self) -> "TimeRange":
        from .timerange import TimeRange  # noqa: F811

        return TimeRange.from_single_timestamp(self)

    @classmethod
    def get_time(cls, *, force_pure_python=False) -> "Timestamp":
        """The force_pure_python keyword only argument is ignored."""
        utc_time = time.time()
        return cls.from_utc(int(utc_time), int(utc_time*cls.MAX_NANOSEC) - int(utc_time)*cls.MAX_NANOSEC)

    @classmethod
    def from_timeoffset(cls, toff: TimeOffsetConstructionType) -> "Timestamp":
        toff = mediatimeoffset(toff)
        return cls(sec=toff.sec, ns=toff.ns, sign=toff.sign)

    @classmethod
    def from_sec_frac(cls, ts_str: str) -> "Timestamp":
        return cast(Timestamp, super(Timestamp, cls).from_sec_frac(ts_str))

    @classmethod
    def from_tai_sec_frac(cls, ts_str: str) -> "Timestamp":
        return cls.from_sec_frac(ts_str)

    @classmethod
    def from_sec_nsec(cls, ts_str: str) -> "Timestamp":
        return cast(Timestamp, super(Timestamp, cls).from_sec_nsec(ts_str))

    @classmethod
    def from_tai_sec_nsec(cls, ts_str: str) -> "Timestamp":
        return cls.from_sec_nsec(ts_str)

    @classmethod
    def from_datetime(cls, dt: datetime) -> "Timestamp":
        minTs = datetime.fromtimestamp(0, tz.gettz('UTC'))
        utcdt = dt.astimezone(tz.gettz('UTC'))
        seconds = int((utcdt - minTs).total_seconds())
        nanoseconds = utcdt.microsecond * 1000

        return cls.from_utc(seconds, nanoseconds, False)

    @classmethod
    def from_iso8601_utc(cls, iso8601utc: str) -> "Timestamp":
        if not iso8601utc.endswith('Z'):
            raise TsValueError("missing 'Z' at end of ISO 8601 UTC format")
        year, month, day, hour, minute, second, ns = _parse_iso8601(iso8601utc[:-1])
        gmtuple = (year, month, day, hour, minute, second - (second == 60))
        secs_since_epoch = calendar.timegm(gmtuple)
        return cls.from_utc(secs_since_epoch, ns, (second == 60))

    @classmethod
    def from_smpte_timelabel(cls, timelabel: str) -> "Timestamp":
        r = re.compile(r'(\d+)-(\d+)-(\d+)T(\d+):(\d+):(\d+)F(\d+) (\d+)/(\d+) UTC([-\+])(\d+):(\d+) TAI([-\+])(\d+)')
        m = r.match(timelabel)
        if m is None:
            raise TsValueError("invalid SMPTE Time Label string format")
        groups = m.groups()
        leap_sec = int(int(groups[5]) == 60)
        local_tm_sec = calendar.timegm(time.struct_time((int(groups[0]), int(groups[1]), int(groups[2]),
                                                         int(groups[3]), int(groups[4]), int(groups[5]) - leap_sec,
                                                         0, 0, 0)))
        rate_num = int(groups[7])
        rate_den = int(groups[8])
        utc_sign = 1
        if groups[9] == '-':
            utc_sign = -1
        tai_sign = 1
        if groups[12] == '-':
            tai_sign = -1
        tai_seconds = (local_tm_sec +
                       leap_sec -
                       utc_sign*(int(groups[10])*60*60 + int(groups[11])*60) -
                       tai_sign*int(groups[13]))
        count = Timestamp(tai_seconds, 0).to_count(rate_num, rate_den, cls.ROUND_UP)
        count += int(groups[6])
        return cls.from_count(count, rate_num, rate_den)

    @classmethod
    def from_str(cls, ts_str: str, *, force_pure_python=False) -> "Timestamp":
        """Parse a string as a TimeStamp

        Accepts SMPTE timelabel, ISO 8601 UTC, second:nanosecond and second.fraction formats, along with "now" to mean
        the current time.

        The force_pure_python keyword only argument is ignored.
        """
        if 'F' in ts_str:
            return cls.from_smpte_timelabel(ts_str)
        elif 'T' in ts_str:
            return cls.from_iso8601_utc(ts_str)
        elif ts_str.strip() == 'now':
            return cls.get_time()
        else:
            return cast(Timestamp, super(Timestamp, cls).from_str(ts_str))

    @classmethod
    def from_count(cls, count: int, rate_num: RationalTypes, rate_den: RationalTypes = 1) -> "Timestamp":
        return cast(Timestamp, super(Timestamp, cls).from_count(count, rate_num, rate_den))

    @classmethod
    def from_utc(cls, utc_sec: int, utc_ns: int, is_leap: bool = False) -> "Timestamp":
        leap_sec = 0
        for tbl_sec, tbl_tai_sec_minus_1 in UTC_LEAP:
            if utc_sec >= tbl_sec:
                leap_sec = (tbl_tai_sec_minus_1 + 1) - tbl_sec
                break
        return cls(sec=utc_sec+leap_sec+is_leap, ns=utc_ns)

    def get_leap_seconds(self) -> int:
        """ Get the UTC leaps seconds.
        Returns the number of leap seconds that the timestamp is adjusted by when
        converting to UTC.
        """
        leap_sec = 0
        for utc_sec, tai_sec_minus_1 in UTC_LEAP:
            if self.sec >= tai_sec_minus_1:
                leap_sec = (tai_sec_minus_1 + 1) - utc_sec
                break

        return leap_sec

    def to_tai_sec_nsec(self) -> str:
        return self.to_sec_nsec()

    def to_tai_sec_frac(self, fixed_size: bool = False) -> str:
        return self.to_sec_frac(fixed_size=fixed_size)

    def to_datetime(self) -> datetime:
        sec, nsec, leap = self.to_utc()
        microsecond = int(round(nsec/1000))
        if microsecond > 999999:
            sec += 1
            microsecond = 0
        dt = datetime.fromtimestamp(sec, tz.gettz('UTC'))
        dt = dt.replace(microsecond=microsecond)

        return dt

    def to_utc(self) -> Tuple[int, int, bool]:
        """ Convert to UTC.
        Returns a tuple of (seconds, nanoseconds, is_leap), where `is_leap` is
        `True` when the input time corresponds exactly to a UTC leap second.
        Note that this deliberately returns a tuple, to try and avoid confusion.
        """
        leap_sec = 0
        is_leap = False
        for utc_sec, tai_sec_minus_1 in UTC_LEAP:
            if self.sec >= tai_sec_minus_1:
                leap_sec = (tai_sec_minus_1 + 1) - utc_sec
                is_leap = self.sec == tai_sec_minus_1
                break

        return (self.sec - leap_sec, self.ns, is_leap)

    def to_iso8601_utc(self) -> str:
        """ Get printed representation in ISO8601 format (UTC)
        YYYY-MM-DDThh:mm:ss.s
        where `s` is fractional seconds at nanosecond precision (always 9-chars wide)
        """
        utc_s, utc_ns, is_leap = self.to_utc()
        utc_bd = time.gmtime(utc_s)
        frac_sec = self._get_fractional_seconds(fixed_size=True)
        leap_sec = int(is_leap)
        return '%04d-%02d-%02dT%02d:%02d:%02d.%sZ' % (utc_bd.tm_year,
                                                      utc_bd.tm_mon,
                                                      utc_bd.tm_mday,
                                                      utc_bd.tm_hour,
                                                      utc_bd.tm_min,
                                                      utc_bd.tm_sec + leap_sec,
                                                      frac_sec)

    def to_smpte_timelabel(self,
                           rate_num: RationalTypes,
                           rate_den: RationalTypes = 1,
                           utc_offset: Optional[int] = None) -> str:
        if rate_num <= 0 or rate_den <= 0:
            raise TsValueError("invalid rate")
        rate = Fraction(rate_num, rate_den)
        count = self.to_count(rate)
        normalised_ts = Timestamp.from_count(count, rate)
        tai_seconds = normalised_ts.sec
        count_on_or_after_second = Timestamp(tai_seconds, 0).to_count(rate, rounding=self.ROUND_UP)
        count_within_second = count - count_on_or_after_second

        utc_sec, utc_ns, is_leap = normalised_ts.to_utc()
        leap_sec = int(is_leap)

        if utc_offset is None:
            # calculate local time offset
            utc_offset_sec = time.timezone
            lt = time.localtime(utc_sec)
            if lt.tm_isdst > 0:
                utc_offset_sec += 60*60
        else:
            utc_offset_sec = utc_offset
        utc_offset_sec_abs = abs(utc_offset_sec)
        utc_offset_hour = utc_offset_sec_abs // (60*60)
        utc_offset_min = (utc_offset_sec_abs % (60*60)) // 60
        utc_sign_char = '+'
        if utc_offset_sec < 0:
            utc_sign_char = '-'

        utc_bd = time.gmtime(utc_sec + utc_offset_sec)

        tai_offset = utc_sec + leap_sec - tai_seconds
        tai_sign_char = '+'
        if tai_offset < 0:
            tai_sign_char = '-'

        return '%04d-%02d-%02dT%02d:%02d:%02dF%02u %u/%u UTC%c%02u:%02u TAI%c%u' % (
                    utc_bd.tm_year, utc_bd.tm_mon, utc_bd.tm_mday,
                    utc_bd.tm_hour, utc_bd.tm_min, utc_bd.tm_sec + leap_sec,
                    count_within_second,
                    rate.numerator, rate.denominator,
                    utc_sign_char, utc_offset_hour, utc_offset_min,
                    tai_sign_char, abs(tai_offset))

    def normalise(self,
                  rate_num: RationalTypes,
                  rate_den: RationalTypes = 1,
                  rounding: "TimeOffset.Rounding" = TimeOffset.ROUND_NEAREST) -> "Timestamp":
        """Return the nearest Timestamp to self which represents an integer number of samples at the given rate.

        :param rate_num: Rate numerator
        :param rate_den: Rate denominator
        :param rounding: How to round, if set to TimeOffset.ROUND_DOWN (resp. TimeOffset.ROUND_UP) this method will only
                         return a TimeOffset less than or equal to this one (resp. greater than or equal to).
        """
        return self.from_count(self.to_count(rate_num, rate_den, rounding), rate_num, rate_den)

    def __add__(self, other_in: TimeOffsetConstructionType) -> "Timestamp":
        return cast(Timestamp, super().__add__(other_in))

    def __iadd__(self, other_in: TimeOffsetConstructionType) -> "Timestamp":
        return cast(Timestamp, super().__add__(other_in))

    def __isub__(self, other_in: TimeOffsetConstructionType) -> "Timestamp":
        return cast(Timestamp, super().__isub__(other_in))

    def __mul__(self, anint: int) -> "Timestamp":
        return cast(Timestamp, super().__mul__(anint))

    def __rmul__(self, anint: int) -> "Timestamp":
        return cast(Timestamp, super().__rmul__(anint))



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fractions import Fraction
from typing import Tuple, Union, Type, TYPE_CHECKING
from typing_extensions import Protocol, runtime_checkable
from abc import ABCMeta, abstractmethod

from ..constants import MAX_NANOSEC, MAX_SECONDS
from ..exceptions import TsValueError

from ._parse import _parse_seconds_fraction
from ._types import RationalTypes

__all__ = ["TimeOffset", "TimeOffsetConstructionType", "SupportsMediaTimeOffset", "mediatimeoffset"]


TimeOffsetConstructionType = Union["TimeOffset", "SupportsMediaTimeOffset", int, float]


if TYPE_CHECKING:
    @runtime_checkable
    class SupportsMediaTimeOffset (Protocol):
        def __mediatimeoffset__(self) -> "TimeOffset":
            ...
else:
    class SupportsMediaTimeOffset (metaclass=ABCMeta):
        """This is an abstract base class for any class that can be automagically converted into a TimeOffset.

        To implement this simply implement the __mediatimeoffset__ magic method. No need to inherit from this
        class explicitly.
        """
        @classmethod
        def __subclasshook__(cls, subclass: Type) -> bool:
            if hasattr(subclass, "__mediatimeoffset__") or hasattr(subclass, "__mediatimestamp__"):
                return True
            else:
                return False

        @abstractmethod
        def __mediatimeoffset__(self) -> "TimeOffset":
            ...


def mediatimeoffset(v: TimeOffsetConstructionType) -> "TimeOffset":
    """This method can be called on any object which supports the __mediatimeoffset__ magic method
    and also on a TimeOffset, an int or a float. It will always return a TimeOffset or raise a ValueError.
    """
    if isinstance(v, TimeOffset):
        return v
    elif isinstance(v, int):
        return TimeOffset(v)
    elif isinstance(v, float):
        return TimeOffset.from_sec_frac(str(v))
    elif hasattr(v, "__mediatimeoffset__"):
        return v.__mediatimeoffset__()
    elif hasattr(v, "__mediatimestamp__"):
        # Since a Timestamp is a TimeOffset we can fall back to this if available
        return v.__mediatimestamp__()  # type: ignore
    else:
        raise ValueError("{!r} cannot be converted to a mediatimestamp.TimeOffset".format(v))


class TimeOffset(object):
    """A nanosecond precision immutable time difference object.

    Note that the canonical representation of a TimeOffset is seconds:nanoseconds, e.g. "4:500000000".
    TimeOffsets in seconds.fractions format (e.g. "4.5") can be parsed, but should not be used for serialization or
    storage due to difficulty disambiguating them from floats.
    """
    class Rounding (int):
        pass

    ROUND_DOWN = Rounding(0)
    ROUND_NEAREST = Rounding(1)
    ROUND_UP = Rounding(2)

    MAX_NANOSEC = MAX_NANOSEC
    MAX_SECONDS = MAX_SECONDS

    def __init__(self, sec: int = 0, ns: int = 0, sign: int = 1):
        (sec, ns, sign) = self._make_valid(int(sec), int(ns), int(sign))

        self.sec: int
        self.ns: int
        self.sign: int

        self.__dict__['sec'] = int(sec)
        self.__dict__['ns'] = int(ns)
        self.__dict__['sign'] = int(sign)

    def __setattr__(self, name: str, value: object) -> None:
        raise TsValueError("Cannot assign to an immutable TimeOffset")

    def __mediatimeoffset__(self) -> "TimeOffset":
        return self

    @classmethod
    def from_timeoffset(cls, toff: TimeOffsetConstructionType) -> "TimeOffset":
        toff = mediatimeoffset(toff)
        return cls(sec=toff.sec, ns=toff.ns, sign=toff.sign)

    @classmethod
    def get_interval_fraction(cls,
                              rate_num: RationalTypes,
                              rate_den: RationalTypes = 1,
                              factor: int = 1) -> "TimeOffset":
        if rate_num <= 0 or rate_den <= 0:
            raise TsValueError("invalid rate")
        if factor < 1:
            raise TsValueError("invalid interval factor")

        rate = Fraction(rate_num, rate_den)
        sec = rate.denominator // (rate.numerator * factor)
        rem = rate.denominator % (rate.numerator * factor)
        ns = cls.MAX_NANOSEC * rem // (rate.numerator * factor)
        return cls(sec=sec, ns=ns)

    @classmethod
    def from_sec_frac(cls, toff_str: str) -> "TimeOffset":
        sec_frac = toff_str.split(".")
        if len(sec_frac) != 1 and len(sec_frac) != 2:
            raise TsValueError("invalid second.fraction format")
        sec = int(sec_frac[0])
        sign = 1
        if sec_frac[0].startswith("-"):
            sign = -1
            sec = abs(sec)
        ns = 0
        if len(sec_frac) > 1:
            ns = _parse_seconds_fraction(sec_frac[1])
        return cls(sec=sec, ns=ns, sign=sign)

    @classmethod
    def from_sec_nsec(cls, toff_str: str) -> "TimeOffset":
        sec_frac = toff_str.split(":")
        if len(sec_frac) != 1 and len(sec_frac) != 2:
            raise TsValueError("invalid second:nanosecond format")
        sec = int(sec_frac[0])
        sign = 1
        if sec_frac[0].startswith("-"):
            sign = -1
            sec = abs(sec)
        ns = 0
        if len(sec_frac) > 1:
            ns = int(sec_frac[1])
        return cls(sec=sec, ns=ns, sign=sign)

    @classmethod
    def from_str(cls, toff_str: str) -> "TimeOffset":
        """Parse a string as a TimeOffset

        Accepts both second:nanosecond and second.fraction formats.
        """
        if '.' in toff_str:
            return cls.from_sec_frac(toff_str)
        else:
            return cls.from_sec_nsec(toff_str)

    @classmethod
    def from_count(cls, count: int, rate_num: RationalTypes, rate_den: RationalTypes = 1) -> "TimeOffset":
        """Returns a new TimeOffset derived from a count and a particular rate.

        :param count: The sample count
        :param rate_num: The numerator of the rate, in Hz
        :param rate_den: The denominator of the rate in Hz
        """
        if rate_num <= 0 or rate_den <= 0:
            raise TsValueError("invalid rate")
        rate = Fraction(rate_num, rate_den)
        abs_count = abs(count)
        sec = (abs_count * rate.denominator) // rate.numerator
        rem = (abs_count * rate.denominator) % rate.numerator
        ns = (rem * cls.MAX_NANOSEC) // rate.numerator
        sign = 1
        if count < 0:
            sign = -1
        return cls(sec=sec, ns=ns, sign=sign)

    @classmethod
    def from_millisec(cls, millisec: int) -> "TimeOffset":
        abs_millisec = abs(millisec)
        sec = abs_millisec // 1000
        ns = (abs_millisec % 1000) * 1000000
        sign = 1
        if millisec < 0:
            sign = -1
        return cls(sec=sec, ns=ns, sign=sign)

    @classmethod
    def from_microsec(cls, microsec: int) -> "TimeOffset":
        abs_microsec = abs(microsec)
        sec = abs_microsec // 1000000
        ns = (abs_microsec % 1000000) * 1000
        sign = 1
        if microsec < 0:
            sign = -1
        return cls(sec=sec, ns=ns, sign=sign)

    @classmethod
    def from_nanosec(cls, nanosec: int) -> "TimeOffset":
        abs_nanosec = abs(nanosec)
        sec = abs_nanosec // cls.MAX_NANOSEC
        ns = abs_nanosec % cls.MAX_NANOSEC
        sign = 1
        if nanosec < 0:
            sign = -1
        return cls(sec=sec, ns=ns, sign=sign)

    def is_null(self) -> bool:
        return self.sec == 0 and self.ns == 0

    def to_sec_nsec(self) -> str:
        """ Convert to <seconds>:<nanoseconds>
        """
        strSign = ""
        if self.sign < 0:
            strSign = "-"
        return u"{}{}:{}".format(strSign, self.sec, self.ns)

    def to_sec_frac(self, fixed_size: bool = False) -> str:
        """ Convert to <seconds>.<fraction>
        """
        strSign = ""
        if self.sign < 0:
            strSign = "-"
        return u"{}{}.{}".format(
            strSign,
            self.sec,
            self._get_fractional_seconds(fixed_size=fixed_size))

    def to_count(self, rate_num: RationalTypes, rate_den: RationalTypes = 1,
                 rounding: "TimeOffset.Rounding" = ROUND_NEAREST) -> int:
        """Returns an integer such that if this TimeOffset is equal to an exact number of samples at the given rate
        then this is equal, and otherwise the value is rounded as indicated by the rounding parameter.

        :param rate_num: numerator of rate
        :param rate_den: denominator of rate
        :param rounding: One of TimeOffset.ROUND_NEAREST, TimeOffset.ROUND_UP, or TimeOffset.ROUND_DOWN
        """
        if rate_num <= 0 or rate_den <= 0:
            raise TsValueError("invalid rate")
        rate = Fraction(rate_num, rate_den)
        abs_off = self.__abs__()
        use_rounding = rounding
        if self.sign < 0:
            if use_rounding == self.ROUND_UP:
                use_rounding = self.ROUND_DOWN
            elif use_rounding == self.ROUND_DOWN:
                use_rounding = self.ROUND_UP
        if use_rounding == self.ROUND_NEAREST:
            rnd_off = TimeOffset.get_interval_fraction(rate, factor=2)
        elif use_rounding == self.ROUND_UP:
            rnd_off = TimeOffset.get_interval_fraction(rate, factor=1) - TimeOffset(0, 1)
        else:
            rnd_off = TimeOffset()
        if rnd_off.sign > 0:
            abs_off += rnd_off

        # off_at_rate = (off_sec + off_nsec) / (rate_den / rate_num)
        #             = (off_sec x rate_num / rate_den) + (off_nsec x rate_num) / rate_den)
        #             = {f1} + {f2}
        # reduce {f1} as follows: a*b/c = ((a/c)*c + a%c) * b) / c = (a/c)*b + (a%c)*b/c
        # then combine the f1 and f2 nanosecond values before division by rate_den to avoid loss of precision when
        # off_sec and off_nsec are not multiples of rate_den, but (off_sec + off_nsec) is
        f1_whole = (abs_off.sec // rate.denominator) * rate.numerator
        f1_dennsec = (abs_off.sec % rate.denominator) * rate.numerator * self.MAX_NANOSEC
        f2_dennsec = abs_off.ns * rate.numerator
        return self.sign * (f1_whole + (f1_dennsec + f2_dennsec) // (rate.denominator * self.MAX_NANOSEC))

    def to_phase_offset(self, rate_num: RationalTypes, rate_den: RationalTypes = 1) -> "TimeOffset":
        """Return the smallest positive TimeOffset such that abs(self - returnval) represents an integer number of
        samples at the given rate"""
        return self - self.normalise(rate_num, rate_den, rounding=TimeOffset.ROUND_DOWN)

    def to_millisec(self, rounding: "TimeOffset.Rounding" = ROUND_NEAREST) -> int:
        use_rounding = rounding
        if self.sign < 0:
            if use_rounding == self.ROUND_UP:
                use_rounding = self.ROUND_DOWN
            elif use_rounding == self.ROUND_DOWN:
                use_rounding = self.ROUND_UP
        round_ns = 0
        if use_rounding == self.ROUND_NEAREST:
            round_ns = 1000000 // 2
        elif use_rounding == self.ROUND_UP:
            round_ns = 1000000 - 1
        return self.sign * (self.sec*1000 + (self.ns + round_ns)//1000000)

    def to_microsec(self, rounding: "TimeOffset.Rounding" = ROUND_NEAREST) -> int:
        use_rounding = rounding
        if self.sign < 0:
            if use_rounding == self.ROUND_UP:
                use_rounding = self.ROUND_DOWN
            elif use_rounding == self.ROUND_DOWN:
                use_rounding = self.ROUND_UP
        round_ns = 0
        if use_rounding == self.ROUND_NEAREST:
            round_ns = 1000 // 2
        elif use_rounding == self.ROUND_UP:
            round_ns = 1000 - 1
        return self.sign * (self.sec*1000000 + (self.ns + round_ns)//1000)

    def to_nanosec(self) -> int:
        return self.sign * (self.sec*self.MAX_NANOSEC + self.ns)

    def normalise(self,
                  rate_num: RationalTypes,
                  rate_den: RationalTypes = 1,
                  rounding: "TimeOffset.Rounding" = ROUND_NEAREST) -> "TimeOffset":
        """Return the nearest TimeOffset to self which represents an integer number of samples at the given rate.

        :param rate_num: Rate numerator
        :param rate_den: Rate denominator
        :param rounding: How to round, if set to TimeOffset.ROUND_DOWN (resp. TimeOffset.ROUND_UP) this method will only
                         return a TimeOffset less than or equal to this one (resp. greater than or equal to).
        """
        return self.from_count(self.to_count(rate_num, rate_den, rounding), rate_num, rate_den)

    def compare(self, other_in: TimeOffsetConstructionType) -> int:
        other = mediatimeoffset(other_in)
        if self.sign != other.sign:
            return self.sign
        elif self.sec < other.sec:
            return -self.sign
        elif self.sec > other.sec:
            return self.sign
        elif self.ns < other.ns:
            return -self.sign
        elif self.ns > other.ns:
            return self.sign
        else:
            return 0

    def __str__(self) -> str:
        return self.to_sec_nsec()

    def __repr__(self) -> str:
        return "{}.from_sec_nsec({!r})".format("mediatimestamp.immutable." + type(self).__name__, self.to_sec_nsec())

    def __abs__(self) -> "TimeOffset":
        return TimeOffset(self.sec, self.ns, 1)

    def __hash__(self) -> int:
        return self.to_nanosec()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, (int, float, TimeOffset)) and self.compare(other) == 0

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    def __lt__(self, other: TimeOffsetConstructionType) -> bool:
        return self.compare(other) < 0

    def __le__(self, other: TimeOffsetConstructionType) -> bool:
        return self.compare(other) <= 0

    def __gt__(self, other: TimeOffsetConstructionType) -> bool:
        return self.compare(other) > 0

    def __ge__(self, other: TimeOffsetConstructionType) -> bool:
        return self.compare(other) >= 0

    def __add__(self, other_in: TimeOffsetConstructionType) -> "TimeOffset":
        from .timestamp import Timestamp

        other = mediatimeoffset(other_in)
        sec = self.sign*self.sec + other.sign*other.sec
        ns = self.sign*self.ns + other.sign*other.ns

        if not isinstance(self, Timestamp) and not isinstance(other, Timestamp):
            return TimeOffset(sec, ns)
        else:
            return Timestamp(sec, ns)

    def __sub__(self, other_in: TimeOffsetConstructionType) -> "TimeOffset":
        from .timestamp import Timestamp

        other = mediatimeoffset(other_in)
        sec = self.sign*self.sec - other.sign*other.sec
        ns = self.sign*self.ns - other.sign*other.ns

        if isinstance(self, Timestamp) and not isinstance(other, Timestamp):
            return Timestamp(sec, ns)
        else:
            return TimeOffset(sec, ns)

    def __iadd__(self, other_in: TimeOffsetConstructionType) -> "TimeOffset":
        other = mediatimeoffset(other_in)
        tmp = self + other
        return self.__class__(tmp.sec, tmp.ns, tmp.sign)

    def __isub__(self, other_in: TimeOffsetConstructionType) -> "TimeOffset":
        other = mediatimeoffset(other_in)
        tmp = self - other
        return self.__class__(tmp.sec, tmp.ns, tmp.sign)

    def __mul__(self, anint: int) -> "TimeOffset":
        sec = self.sec * abs(anint)
        ns = self.ns * abs(anint)

        if anint < 0:
            sign = self.sign * -1
        else:
            sign = self.sign

        if ns >= self.MAX_NANOSEC:
            sec += (ns // self.MAX_NANOSEC)
            ns %= self.MAX_NANOSEC

        return TimeOffset(sec, ns, sign)

    def __rmul__(self, anint: int) -> "TimeOffset":
        return (self * anint)

    def __div__(self, anint: int) -> "TimeOffset":
        return (self // anint)

    def __truediv__(self, anint: int) -> "TimeOffset":
        return (self // anint)

    def __floordiv__(self, anint: int) -> "TimeOffset":
        (sec, ns, sign) = (self.sec, self.ns, self.sign)
        abs_anint = abs(anint)
        sec = sec // abs_anint
        ns = int((self.ns + (self.sec % abs_anint) * self.MAX_NANOSEC) / abs_anint + 5e-10)

        sec = sec + ns // self.MAX_NANOSEC
        ns = ns % self.MAX_NANOSEC
        if anint < 0:
            sign *= -1
        return TimeOffset(sec, ns, sign)

    def _get_fractional_seconds(self, fixed_size: bool = False) -> str:
        div = self.MAX_NANOSEC // 10
        rem = self.ns
        sec_frac = ""

        for i in range(0, 9):
            if not fixed_size and i > 0 and rem == 0:
                break

            sec_frac += '%i' % (rem / div)
            rem %= div
            div //= 10

        return sec_frac

    def _make_valid(self, sec: int, ns: int, sign: int) -> Tuple[int, int, int]:
        if sign > 0 or (sec == 0 and ns == 0):
            sign = 1
        else:
            sign = -1

        sec += (ns // self.MAX_NANOSEC)
        ns %= self.MAX_NANOSEC

        if sec < 0 or (sec == 0 and ns < 0):
            sec *= -1
            ns *= -1
            sign *= -1

        if ns < 0:
            sec -= 1
            ns += self.MAX_NANOSEC

        if sec >= self.MAX_SECONDS:
            sec = self.MAX_SECONDS - 1
            ns = self.MAX_NANOSEC - 1

        return (sec, ns, sign)



# Copyright 2019 British Broadcasting Corporation
#
# This is an internal BBC tool and is not licensed externally
# If you have received a copy of this erroneously then you do
# not have permission to reproduce it.

from typing import Optional, Union, Any
from fractions import Fraction

from .immutable import (
    TimeOffset,
    Timestamp,
    TimeRange,
    SupportsMediaTimeOffset,
    SupportsMediaTimestamp,
    mediatimeoffset,
    mediatimestamp)


TimeValueRepTypes = Union[Timestamp, TimeOffset, int]
TimeValueConstructTypes = Union[SupportsMediaTimeOffset, SupportsMediaTimestamp, int, "TimeValue"]


def _perform_all_conversions(v: TimeValueConstructTypes) -> Union[TimeValueRepTypes, "TimeValue"]:
    if isinstance(v, (int, TimeOffset, TimeValue)):
        return v
    elif isinstance(v, SupportsMediaTimestamp):
        return mediatimestamp(v)
    elif isinstance(v, SupportsMediaTimeOffset):
        return mediatimeoffset(v)
    else:
        raise TypeError("{!r} is not a valid construction type for a TimeValue".format(v))


class TimeValue(object):
    """Represents a media unit time value on a timeline (e.g. Flow).

    Supports one of the following input value representations:
    * TimeOffset
    * Timestamp
    * int (integer media unit count)
    * TimeValue
    * Anything that implements the __mediatimestamp__ magic method
    * Anything that implements the __mediatimeoffset__ magic method

    An optional rate can be set and is required and checked when there is a
    need to convert between representations.

    TimeOffset and Timestamps are converted internally to ints if a rate is
    provided.

    The time value can be converted to a TimeOffset, Timestamp or int using
    the as_*() methods.
    """

    def __init__(self, value: TimeValueConstructTypes, rate: Optional[Fraction] = None):
        """
        :param value: A TimeValue, TimeOffset, TimeStamp or int.
        :param rate: The media unit rate.
        """
        self_rate: Optional[Fraction]
        self_value: TimeValueRepTypes

        value = _perform_all_conversions(value)

        if isinstance(value, TimeValue):
            if rate and value._rate != rate:
                # A rate conversion is required. Convert to a timeoffset here and the
                # conversion at the end using the new rate
                try:
                    self_value = value.as_timeoffset()
                except ValueError:
                    # the representation is a count and so we assume it is as the given rate
                    self_value = value._value
                self_rate = rate
            else:
                self_value = value._value
                self_rate = value._rate
        elif isinstance(value, (TimeOffset, int)):
            self_value = value
            self_rate = rate
        else:
            raise TypeError("Unsupported value type {!r}".format(value))

        # Convert to an int if the value is a TimeOffset and a rate is
        # provided. This allows for more efficient calculations and no
        # normalisation is required.
        if isinstance(self_value, TimeOffset) and self_rate:
            self_value = self_value.to_count(self_rate.numerator, self_rate.denominator)

        # set attributes using dict to workaround immutability
        self.__dict__['_value'] = self_value
        self.__dict__['_rate'] = self_rate

        # provide attribute type info given that attributes are not set directly
        self._value: TimeValueRepTypes
        self._rate: Optional[Fraction]

    @classmethod
    def from_str(cls, s: str, rate: Optional[Fraction] = None) -> "TimeValue":
        """Parse a time value string

        :param s: The string to convert from.
        :param rate: The default media unit rate.
        """
        parts = s.split("@")
        if len(parts) == 2:
            s_val = parts[0]
            rate = Fraction(parts[1])
        elif len(parts) == 1:
            s_val = s
        else:
            raise ValueError("Multiple '@' in TimeValue string")

        if s_val.isdigit() or (
                len(s_val) > 0 and s_val[0] in ['+', '-'] and s_val[1:].isdigit()):
            return cls(int(s_val), rate=rate)
        else:
            # Assuming that it represents a TimeOffset rather than a Timestamp
            return cls(TimeOffset.from_str(s_val), rate=rate)

    def as_timeoffset(self) -> TimeOffset:
        """Returns a TimeOffset representation."""
        if isinstance(self._value, Timestamp):
            return TimeOffset.from_timeoffset(self._value)
        elif isinstance(self._value, TimeOffset):
            return self._value
        else:
            rate = self._require_rate()
            return TimeOffset.from_count(self._value, rate.numerator, rate.denominator)

    def __mediatimeoffset__(self) -> TimeOffset:
        return self.as_timeoffset()

    def as_timestamp(self) -> Timestamp:
        """Returns a Timestamp representation."""
        if isinstance(self._value, Timestamp):
            return self._value
        elif isinstance(self._value, TimeOffset):
            return Timestamp.from_timeoffset(self._value)
        else:
            rate = self._require_rate()
            return Timestamp.from_count(self._value, rate.numerator, rate.denominator)

    def __mediatimestamp__(self) -> Timestamp:
        return self.as_timestamp()

    def __mediatimerange__(self) -> TimeRange:
        return TimeRange.from_single_timestamp(self.as_timestamp())

    def as_count(self) -> int:
        """Returns an integer media unit count representation."""
        if isinstance(self._value, TimeOffset):
            rate = self._require_rate()
            return self._value.to_count(rate.numerator, rate.denominator)
        else:
            return self._value

    @property
    def rate(self) -> Optional[Fraction]:
        return self._rate

    @property
    def value(self) -> TimeValueRepTypes:
        return self._value

    def compare(self, other: TimeValueConstructTypes) -> int:
        """Compare time values and return an integer to indicate the difference"""
        other = _perform_all_conversions(other)
        other_value = self._match_value_type(other)
        if isinstance(self._value, TimeOffset):
            return self._value.compare(other_value)
        else:
            # The logic here follows the TimeOffset implementation
            this_sign = 1 if self >= 0 else -1
            other_sign = 1 if other >= 0 else -1
            if this_sign != other_sign:
                return this_sign
            elif self < other:
                return -this_sign
            elif self > other:
                return this_sign
            else:
                return 0

    def to_str(self, include_rate: bool = True) -> str:
        """Convert to a string"""
        result = str(self._value)
        if self._rate and include_rate:
            result += "@{}".format(self._rate)
        return result

    def __setattr__(self, name: str, value: Any) -> None:
        """Raises a ValueError if attempt to set an attribute on the immutable TimeValue"""
        raise ValueError("Cannot assign to an immutable TimeValue")

    def __str__(self) -> str:
        return self.to_str()

    def __repr__(self) -> str:
        return "{}.{}.from_str('{}')".format(type(self).__module__, type(self).__name__, self.to_str())

    def __abs__(self) -> "TimeValue":
        """Return the absolute TimeValue"""
        return TimeValue(self._value.__abs__(), self._rate)

    def __eq__(self, other: object) -> bool:
        """"Return true if the TimeValues are equal"""
        if not isinstance(other, (SupportsMediaTimestamp, SupportsMediaTimeOffset, int, TimeValue)):
            return False
        other_value = self._match_value_type(other)
        return self._value.__eq__(other_value)

    def __ne__(self, other: object) -> bool:
        """"Return true if the TimeValues are not equal"""
        if not isinstance(other, (SupportsMediaTimestamp, SupportsMediaTimeOffset, int, TimeValue)):
            return True
        other_value = self._match_value_type(other)
        return self._value.__ne__(other_value)

    def __lt__(self, other: TimeValueConstructTypes) -> bool:
        """"Return true if this TimeValue is less than the other TimeValue"""
        other_value = self._match_value_type(other)
        return self._value.__lt__(other_value)  # type: ignore

    def __le__(self, other: TimeValueConstructTypes) -> bool:
        """"Return true if this TimeValue is less than or equal to the other TimeValue"""
        other_value = self._match_value_type(other)
        return self._value.__le__(other_value)  # type: ignore

    def __gt__(self, other: TimeValueConstructTypes) -> bool:
        """"Return true if this TimeValue is greater than the other TimeValue"""
        other_value = self._match_value_type(other)
        return self._value.__gt__(other_value)  # type: ignore

    def __ge__(self, other: TimeValueConstructTypes) -> bool:
        """"Return true if this TimeValue is greater than or equal to the other TimeValue"""
        other_value = self._match_value_type(other)
        return self._value.__ge__(other_value)  # type: ignore

    def __add__(self, other: TimeValueConstructTypes) -> "TimeValue":
        """"Return a TimeValue that is the sum of this and the other TimeValue"""
        other_value = self._match_value_type(other)
        return TimeValue(self._value.__add__(other_value), self._rate)  # type: ignore

    def __sub__(self, other: TimeValueConstructTypes) -> "TimeValue":
        """"Return a TimeValue that is the difference between this and the other TimeValue"""
        other_value = self._match_value_type(other)
        return TimeValue(self._value.__sub__(other_value), self._rate)  # type: ignore

    def __mul__(self, anint: int) -> "TimeValue":
        """"Return this TimeValue multiplied by an integer"""
        self._check_is_int(anint)
        return TimeValue(self._value.__mul__(anint), self._rate)

    def __rmul__(self, anint: int) -> "TimeValue":
        """"Return this TimeValue multiplied by an integer"""
        self._check_is_int(anint)
        return (self * anint)

    def __div__(self, anint: int) -> "TimeValue":
        """"Return this TimeValue divided by an integer, rounded down to -inf"""
        self._check_is_int(anint)
        return (self // anint)

    def __truediv__(self, anint: int) -> "TimeValue":
        """"Return this TimeValue divided by an integer, rounded down to -inf"""
        self._check_is_int(anint)
        return (self // anint)

    def __floordiv__(self, anint: int) -> "TimeValue":
        """"Return this TimeValue divided by an integer, rounded down to -inf"""
        self._check_is_int(anint)
        return TimeValue(self._value.__floordiv__(anint), self._rate)

    def __hash__(self) -> int:
        return hash(repr(self))

    def _match_value_type(self, other: TimeValueConstructTypes) -> TimeValueRepTypes:
        """Converts the other value type to self's value type.

        A rate conversion is done if `other` is a TimeValue and the rate
        differs from self._rate.

        :param other: A TimeValue, TimeOffset, Timestamp or int.
        """
        other_tv = TimeValue(other, rate=self._rate)
        if isinstance(self._value, Timestamp):
            return other_tv.as_timestamp()
        elif isinstance(self._value, TimeOffset):
            return other_tv.as_timeoffset()
        else:
            return other_tv.as_count()

    def _require_rate(self) -> Fraction:
        """Raise an exception if the self._rate is not set or zero.
        """
        if not self._rate:
            raise ValueError("A non-zero TimeValue rate is required for conversion")

        return self._rate

    def _check_is_int(self, anint: Any) -> None:
        """Raise an exception if the parameter is not an integer.

        :param anint: Parameter to check.
        """
        if not isinstance(anint, int):
            raise TypeError("TimeValue operator parameter {!r} is not an 'int'".format(anint))



# Copyright 2019 British Broadcasting Corporation
#
# This is an internal BBC tool and is not licensed externally
# If you have received a copy of this erroneously then you do
# not have permission to reproduce it.

from typing import Optional, Union, Any, Tuple
import re


class CountRange(object):
    """Represents a range of integer media unit counts.

    The implementation matches mediatimestamp.immutable.TimeRange, but
    with Timestamps replaced with integer counts.
    """

    # These inclusivity and rounding values must match the TimeRange values

    EXCLUSIVE = 0x0
    INCLUDE_START = 0x1
    INCLUDE_END = 0x2
    INCLUSIVE = 0x3

    ROUND_DOWN = 0
    ROUND_NEAREST = 1
    ROUND_UP = 2
    ROUND_IN = 3
    ROUND_OUT = 4
    ROUND_START = 5
    ROUND_END = 6

    def __init__(self, start: Optional[int],
                 end: Optional[int] = None,
                 inclusivity: int = INCLUSIVE):
        """Construct a count range starting at start and ending at end

        :param start: An integer media unit count or None
        :param end: An integer media unit count or None
        :param inclusivity: a combination of flags INCLUDE_START and INCLUDE_END"""

        # normalise the representation to always have an exclusive end if bounded
        if end is not None and (inclusivity & CountRange.INCLUDE_END):
            end = end + 1
            inclusivity &= ~CountRange.INCLUDE_END

        # normalise the representation to always have an inclusive start if bounded
        if start is not None and not (inclusivity & CountRange.INCLUDE_START):
            start = start + 1
            inclusivity |= CountRange.INCLUDE_START

        # normalise 'never' cases
        if start is not None and end is not None:
            if start > end or (start == end and inclusivity != CountRange.INCLUSIVE):
                start = 0
                end = 0
                inclusivity = CountRange.EXCLUSIVE

        # Normalise the 'eternity' cases
        if start is None and end is None:
            inclusivity = CountRange.INCLUSIVE

        # set attributes using dict to workaround immutability
        self.__dict__['start'] = start
        self.__dict__['end'] = end
        self.__dict__['inclusivity'] = inclusivity

        # provide attribute type info given that attributes are not set directly
        self.start: int
        self.end: int
        self.inclusivity: int

    @classmethod
    def from_start(cls, start: int, inclusivity: int = INCLUSIVE) -> "CountRange":
        """Construct a range starting at start with no end

        :param start: An integer media unit count
        :param inclusivity: a combination of flags INCLUDE_START and INCLUDE_END"""
        return cls(start, None, inclusivity)

    @classmethod
    def from_end(cls, end: int, inclusivity: int = INCLUSIVE) -> "CountRange":
        """Construct a range ending at end with no start

        :param end: An integer media unit count
        :param inclusivity: a combination of flags INCLUDE_START and INCLUDE_END"""
        return cls(None, end, inclusivity)

    @classmethod
    def from_start_length(cls, start: int, length: int, inclusivity: int = INCLUSIVE) -> "CountRange":
        """Construct a range starting at start and ending at (start + length)

        :param start: An integer media unit count
        :param length: An integer media unit offset, which must be non-negative
        :param inclusivity: a combination of flags INCLUDE_START and INCLUDE_END

        :raises: ValueError if the length is negative"""
        if length < 0:
            raise ValueError("Length must be non-negative")
        return cls(start, start + length, inclusivity)

    @classmethod
    def eternity(cls) -> "CountRange":
        """Return an unbounded range covering all time"""
        return cls(None, None)

    @classmethod
    def never(cls) -> "CountRange":
        """Return a range covering no time"""
        return cls(0, 0, CountRange.EXCLUSIVE)

    @classmethod
    def from_single_count(cls, count: int) -> "CountRange":
        """Construct a range containing only a single count

        :param count: An integer media unit count"""
        return cls(count, count, CountRange.INCLUSIVE)

    @classmethod
    def from_str(cls, s: str, inclusivity: int = INCLUSIVE) -> "CountRange":
        """Convert a string to a range.

        Valid ranges are:
        [<count>_<count>]
        [<count>_<count>)
        (<count>_<count>]
        (<count>_<count>)
        [<count>]
        <count>_<count>
        <count>
        ()

        where <count> is an integer or an empty string.

        The meaning of these is relatively simple: [ indicates including the start count,
        ( indicates excluding it, ] indicates including the end count, and ) indicates excluding it.
        If brackets are ommitted entirely then this is taken as an inclusive range at both ends.
        Omitting a count indicates that there is no bound on that end (ie. the range goes on forever),
        including only a single count by itself indicates a range containing exactly that one count.
        Finally the string "()" represents the empty range.

        :param s: The string to process
        """
        m = re.match(r'(\[|\()?([^_\)\]]+)?(_([^_\)\]]+)?)?(\]|\))?', s)

        if m is None:
            raise ValueError("Invalid CountRange string")

        inc = CountRange.INCLUSIVE
        if m.group(1) == "(":
            inc &= ~CountRange.INCLUDE_START
        if m.group(5) == ")":
            inc &= ~CountRange.INCLUDE_END

        start_str = m.group(2)
        end_str = m.group(4)

        if start_str is not None:
            start = int(start_str)
        else:
            start = None
        if end_str is not None:
            end = int(end_str)
        else:
            end = None

        if start is None and end is None:
            # Ie. we have no first or second count
            if m.group(3) is not None:
                # ie. we have a '_' character
                return cls.eternity()
            else:
                # We have no '_' character, so the whole range is empty
                return cls.never()
        elif start is not None and end is None and m.group(3) is None:
            return cls.from_single_count(start)
        else:
            return cls(start, end, inc)

    @property
    def length(self) -> Union[int, float]:
        """Return the range length as a media unit count"""
        if self.end is None or self.start is None:
            return float("inf")  # there is no int("inf") in python
        return self.end - self.start

    def bounded_before(self) -> bool:
        """Return true if the start of the range is bounded"""
        return self.start is not None

    def bounded_after(self) -> bool:
        """Return true if the end of the range is bounded"""
        return self.end is not None

    def unbounded(self) -> bool:
        """Return true if neither the start or end of the range is bounded"""
        return self.start is None and self.end is None

    def includes_start(self) -> bool:
        """Return true if the start is inclusive"""
        return (self.inclusivity & CountRange.INCLUDE_START) != 0

    def includes_end(self) -> bool:
        """Return true if the end is inclusive"""
        return (self.inclusivity & CountRange.INCLUDE_END) != 0

    def finite(self) -> bool:
        """Return true if the range is finite"""
        return (self.start is not None and self.end is not None)

    def contains_subrange(self, other: "CountRange") -> bool:
        """Returns true if the range supplied lies entirely inside this range"""
        return ((not self.is_empty()) and
                (other.is_empty() or
                 (self.start is None or (other.start is not None and self.start <= other.start)) and
                 (self.end is None or (other.end is not None and self.end >= other.end)) and
                 (not ((self.start is not None) and
                       (other.start is not None) and
                       (self.start == other.start) and
                       (self.inclusivity & CountRange.INCLUDE_START == 0) and
                       (other.inclusivity & CountRange.INCLUDE_START != 0))) and
                 (not ((self.end is not None) and
                       (other.end is not None) and
                       (self.end == other.end) and
                       (self.inclusivity & CountRange.INCLUDE_END == 0) and
                       (other.inclusivity & CountRange.INCLUDE_END != 0)))))

    def to_str(self, with_inclusivity_markers: bool = True) -> str:
        """Convert to [<count>_<count>] format,
        usually the opening and closing delimiters are set to [ or ] for inclusive and ( or ) for exclusive ranges.
        Unbounded ranges have no marker attached to them.

        :param with_inclusivity_markers: if set to False do not include parentheses/brackecount"""
        if self.is_empty():
            if with_inclusivity_markers:
                return "()"
            else:
                return ""
        elif self.start is not None and self.end is not None and self.start == self.end:
            if with_inclusivity_markers:
                return "[" + str(self.start) + "]"
            else:
                return str(self.start)

        if with_inclusivity_markers:
            brackets = [("(", ")"), ("[", ")"), ("(", "]"), ("[", "]")][self.inclusivity]
        else:
            brackets = ("", "")

        return '_'.join([
            (brackets[0] + str(self.start)) if self.start is not None else '',
            (str(self.end) + brackets[1]) if self.end is not None else ''
            ])

    def intersect_with(self, other: "CountRange") -> "CountRange":
        """Return a range which represents the intersection of this range with another"""
        if self.is_empty() or other.is_empty():
            return CountRange.never()

        start = self.start
        if other.start is not None and (self.start is None or self.start < other.start):
            start = other.start
        end = self.end
        if other.end is not None and (self.end is None or self.end > other.end):
            end = other.end

        inclusivity = CountRange.EXCLUSIVE
        if start is None or (start in self and start in other):
            inclusivity |= CountRange.INCLUDE_START
        if end is None or (end in self and end in other):
            inclusivity |= CountRange.INCLUDE_END

        if start is not None and end is not None and start > end:
            return CountRange.never()

        return CountRange(start, end, inclusivity)

    def starts_inside_range(self, other: "CountRange") -> bool:
        """Returns true if the start of this range is located inside the other."""
        return (not self.is_empty() and
                not other.is_empty() and
                ((self.bounded_before() and self.start in other and
                  (not (other.bounded_after() and self.start == other.end and not self.includes_start()))) or
                 (self.bounded_before() and other.bounded_before() and self.start == other.start and
                  (not (self.includes_start() and not other.includes_start()))) or
                 (not self.bounded_before() and not other.bounded_before())))

    def ends_inside_range(self, other: "CountRange") -> bool:
        """Returns true if the end of this range is located inside the other."""
        return (not self.is_empty() and
                not other.is_empty() and
                ((self.bounded_after() and self.end in other and
                  (not (other.bounded_before() and self.end == other.start and not self.includes_end()))) or
                 (self.bounded_after() and other.bounded_after() and self.end == other.end and
                  (not (self.includes_end() and not other.includes_end()))) or
                 (not self.bounded_after() and not other.bounded_after())))

    def is_earlier_than_range(self, other: "CountRange") -> bool:
        """Returns true if this range ends earlier than the start of the other."""
        return (not self.is_empty() and
                not other.is_empty() and
                other.bounded_before() and
                self.bounded_after() and
                (self.end < other.start or
                 (self.end == other.start and
                  not (self.includes_end() and other.includes_start()))))

    def is_later_than_range(self, other: "CountRange") -> bool:
        """Returns true if this range starts later than the end of the other."""
        return (not self.is_empty() and
                not other.is_empty() and
                other.bounded_after() and
                self.bounded_before() and
                (self.start > other.end or
                 (self.start == other.end and
                  not (self.includes_start() and other.includes_end()))))

    def starts_earlier_than_range(self, other: "CountRange") -> bool:
        """Returns true if this range starts earlier than the start of the other."""
        return (not self.is_empty() and
                not other.is_empty() and
                other.bounded_before() and
                (not self.bounded_before() or
                 (self.start < other.start or
                  (self.start == other.start and
                   self.includes_start() and
                   not other.includes_start()))))

    def starts_later_than_range(self, other: "CountRange") -> bool:
        """Returns true if this range starts later than the start of the other."""
        return (not self.is_empty() and
                not other.is_empty() and
                self.bounded_before() and
                (not other.bounded_before() or
                 (self.start > other.start or
                  (self.start == other.start and
                   (not self.includes_start() and other.includes_start())))))

    def ends_earlier_than_range(self, other: "CountRange") -> bool:
        """Returns true if this range ends earlier than the end of the other."""
        return (not self.is_empty() and
                not other.is_empty() and
                self.bounded_after() and
                (not other.bounded_after() or
                 (self.end < other.end or
                  (self.end == other.end and
                   (not self.includes_end() and other.includes_end())))))

    def ends_later_than_range(self, other: "CountRange") -> bool:
        """Returns true if this range ends later than the end of the other."""
        return (not self.is_empty() and
                not other.is_empty() and
                other.bounded_after() and
                (not self.bounded_after() or
                 (self.end > other.end or
                  (self.end == other.end and
                   self.includes_end() and
                   not other.includes_end()))))

    def overlaps_with_range(self, other: "CountRange") -> bool:
        """Returns true if this range and the other overlap."""
        return (not self.is_earlier_than_range(other) and not self.is_later_than_range(other))

    def is_contiguous_with_range(self, other: "CountRange") -> bool:
        """Returns true if the union of this range and the other would be a valid range"""
        return (self.overlaps_with_range(other) or
                (self.is_earlier_than_range(other) and
                 self.end == other.start and
                 (self.includes_end() or other.includes_start())) or
                (self.is_later_than_range(other) and
                 self.start == other.end and
                 (self.includes_start() or other.includes_end())))

    def union_with_range(self, other: "CountRange") -> "CountRange":
        """Returns the union of this range and the other.
        :raises: ValueError if the ranges are not contiguous."""
        if not self.is_contiguous_with_range(other):
            raise ValueError("CountRanges {} and {} are not contiguous, so cannot take the union.".format(self, other))

        return self.extend_to_encompass_range(other)

    def extend_to_encompass_range(self, other: "CountRange") -> "CountRange":
        """Returns the range that encompasses this and the other range."""
        if self.is_empty():
            return other

        if other.is_empty():
            return self

        inclusivity = CountRange.EXCLUSIVE
        if self.start == other.start:
            start = self.start
            inclusivity |= ((self.inclusivity | other.inclusivity) & CountRange.INCLUDE_START)
        elif self.starts_earlier_than_range(other):
            start = self.start
            inclusivity |= (self.inclusivity & CountRange.INCLUDE_START)
        else:
            start = other.start
            inclusivity |= (other.inclusivity & CountRange.INCLUDE_START)

        if self.end == other.end:
            end = self.end
            inclusivity |= ((self.inclusivity | other.inclusivity) & CountRange.INCLUDE_END)
        elif self.ends_later_than_range(other):
            end = self.end
            inclusivity |= (self.inclusivity & CountRange.INCLUDE_END)
        else:
            end = other.end
            inclusivity |= (other.inclusivity & CountRange.INCLUDE_END)

        return CountRange(start, end, inclusivity)

    def split_at(self, count: int) -> Tuple["CountRange", "CountRange"]:
        """Splits a range at a specified count.

        It is guaranteed that the splitting point will be in the *second* CountRange returned, and not in the first.

        :param count: the count to split at
        :returns: A pair of CountRange objects
        :raises: ValueError if count not in self"""

        if count not in self:
            raise ValueError("Cannot split range {} at {}".format(self, count))

        return (CountRange(self.start, count, (self.inclusivity & CountRange.INCLUDE_START)),
                CountRange(count, self.end, CountRange.INCLUDE_START | (self.inclusivity & CountRange.INCLUDE_END)))

    def range_between(self, other: "CountRange") -> "CountRange":
        """Returns the range between the end of the earlier range and the start of the later one"""
        if self.is_contiguous_with_range(other):
            return CountRange.never()
        elif self.is_earlier_than_range(other):
            inclusivity = CountRange.EXCLUSIVE
            if not self.includes_end():
                inclusivity |= CountRange.INCLUDE_START
            if not other.includes_start():
                inclusivity |= CountRange.INCLUDE_END
            return CountRange(self.end, other.start, inclusivity)
        else:
            inclusivity = CountRange.EXCLUSIVE
            if not self.includes_start():
                inclusivity |= CountRange.INCLUDE_END
            if not other.includes_end():
                inclusivity |= CountRange.INCLUDE_START
            return CountRange(other.end, self.start, inclusivity)

    def is_empty(self) -> bool:
        """Returns true on any empty range."""
        return (self.start is not None and
                self.end is not None and
                self.start == self.end and
                self.inclusivity != CountRange.INCLUSIVE)

    def __setattr__(self, name: str, value: Any) -> None:
        """Raises a ValueError if attempt to set an attribute on the immutable CountRange"""
        raise ValueError("Cannot assign to an immutable CountRange")

    def __contains__(self, count: int) -> bool:
        """Returns true if the count is within this range."""
        return ((self.start is None or count >= self.start) and
                (self.end is None or count <= self.end) and
                (not ((self.start is not None) and
                      (count == self.start) and
                      (self.inclusivity & CountRange.INCLUDE_START == 0))) and
                (not ((self.end is not None) and
                      (count == self.end) and
                      (self.inclusivity & CountRange.INCLUDE_END == 0))))

    def __eq__(self, other: Any) -> bool:
        """Return true if the ranges are equal"""
        return (isinstance(other, CountRange) and
                ((self.is_empty() and other.is_empty()) or
                (((self.start is None and other.start is None) or
                  (self.start == other.start and
                   (self.inclusivity & CountRange.INCLUDE_START) == (other.inclusivity & CountRange.INCLUDE_START))) and
                 ((self.end is None and other.end is None) or
                  (self.end == other.end and
                   (self.inclusivity & CountRange.INCLUDE_END) == (other.inclusivity & CountRange.INCLUDE_END))))))

    def __str__(self) -> str:
        return self.to_str()

    def __repr__(self) -> str:
        return "{}.{}.from_str('{}')".format(type(self).__module__, type(self).__name__, self.to_str())



# Copyright 2020 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from fractions import Fraction

from copy import deepcopy

from mediatimestamp.immutable import TimeOffset, TsValueError, SupportsMediaTimeOffset, mediatimeoffset


class TestTimeOffset(unittest.TestCase):
    def test_supportsmediatimeoffset(self):
        to = TimeOffset()
        self.assertIsInstance(to, SupportsMediaTimeOffset)

        class _convertable(object):
            def __mediatimeoffset__(self) -> TimeOffset:
                return TimeOffset()

        c = _convertable()
        self.assertIsInstance(c, SupportsMediaTimeOffset)

        self.assertEqual(to, mediatimeoffset(to))
        self.assertEqual(to, mediatimeoffset(c))

    def test_MAX_NANOSEC(self):
        self.assertEqual(TimeOffset.MAX_NANOSEC, 1000000000)

    def test_from_timeoffset(self):
        """This tests that TimeOffsets can be created with a variety of values."""
        tests_ts = [
            (TimeOffset.from_timeoffset(TimeOffset(0, 0)), TimeOffset(0, 0)),
            (TimeOffset.from_timeoffset(TimeOffset(1001, 0)), TimeOffset(1001, 0)),
            (TimeOffset.from_timeoffset(TimeOffset(1001, 1001)), TimeOffset(1001, 1001)),
            ]

        for t in tests_ts:
            with self.subTest(t=t):
                self.assertEqual(t[0], t[1])

    def test_normalise(self):
        tests_ts = [
            (TimeOffset(0, 0), Fraction(30000, 1001), TimeOffset.ROUND_NEAREST,
             TimeOffset(0, 0)),
            (TimeOffset(1001, 0), Fraction(30000, 1001), TimeOffset.ROUND_NEAREST,
             TimeOffset(1001, 0)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000), Fraction(30000, 1001), TimeOffset.ROUND_NEAREST,
             TimeOffset(1001, 0)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000 + 1), Fraction(30000, 1001), TimeOffset.ROUND_NEAREST,
             TimeOffset(1001, 1001.0/30000*1000000000)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000, -1), Fraction(30000, 1001), TimeOffset.ROUND_NEAREST,
             TimeOffset(1001, 0, -1)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000 + 1, -1), Fraction(30000, 1001), TimeOffset.ROUND_NEAREST,
             TimeOffset(1001, 1001.0/30000*1000000000, -1)),
            (TimeOffset(1521731233, 320000000), Fraction(25, 3), TimeOffset.ROUND_NEAREST,
             TimeOffset(1521731233, 320000000)),
            (TimeOffset(0, 0), Fraction(30000, 1001), TimeOffset.ROUND_UP,
             TimeOffset(0, 0)),
            (TimeOffset(1001, 0), Fraction(30000, 1001), TimeOffset.ROUND_UP,
             TimeOffset(1001, 0)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000), Fraction(30000, 1001), TimeOffset.ROUND_UP,
             TimeOffset(1001, 1001.0/30000*1000000000)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000 + 1), Fraction(30000, 1001), TimeOffset.ROUND_UP,
             TimeOffset(1001, 1001.0/30000*1000000000)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000, -1), Fraction(30000, 1001), TimeOffset.ROUND_UP,
             TimeOffset(1001, 0, -1)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000 + 1, -1), Fraction(30000, 1001), TimeOffset.ROUND_UP,
             TimeOffset(1001, 0, -1)),
            (TimeOffset(1521731233, 320000000), Fraction(25, 3), TimeOffset.ROUND_UP,
             TimeOffset(1521731233, 320000000)),
            (TimeOffset(0, 0), Fraction(30000, 1001), TimeOffset.ROUND_DOWN,
             TimeOffset(0, 0)),
            (TimeOffset(1001, 0), Fraction(30000, 1001), TimeOffset.ROUND_DOWN,
             TimeOffset(1001, 0)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000), Fraction(30000, 1001), TimeOffset.ROUND_DOWN,
             TimeOffset(1001, 0)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000 + 1), Fraction(30000, 1001), TimeOffset.ROUND_DOWN,
             TimeOffset(1001, 0)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000, -1), Fraction(30000, 1001), TimeOffset.ROUND_DOWN,
             TimeOffset(1001, 1001.0/30000*1000000000, -1)),
            (TimeOffset(1001, 1001.0/30000/2*1000000000 + 1, -1), Fraction(30000, 1001), TimeOffset.ROUND_DOWN,
             TimeOffset(1001, 1001.0/30000*1000000000, -1)),
            (TimeOffset(1521731233, 320000000), Fraction(25, 3), TimeOffset.ROUND_DOWN,
             TimeOffset(1521731233, 320000000)),
        ]

        n = 0
        for (input, rate, rounding, expected) in tests_ts:
            # Nb. subTest will add a printout of all its kwargs to any error message generated
            # by a failure within it. The variable n is being used here to ensure that the index
            # of the current test within tests_ts is printed on any failure. (Nb. only works with
            # python3 unittest test runner)
            with self.subTest(test_data_index=n,
                              input=input,
                              rate=rate,
                              rounding=rounding,
                              expected=expected):
                n += 1
                r = input.normalise(rate.numerator, rate.denominator, rounding=rounding)
                self.assertEqual(r, expected,
                                 msg=("{!r}.normalise({}, {}, rounding={}) == {!r}, expected {!r}"
                                      .format(input, rate.numerator, rate.denominator, rounding, r, expected)))

    def test_hash(self):
        self.assertEqual(hash(TimeOffset(0, 0)), hash(TimeOffset(0, 0)))
        self.assertNotEqual(hash(TimeOffset(0, 0)), hash(TimeOffset(0, 1)))

    def test_subsec(self):
        """This tests that TimeOffsets can be converted to millisec, nanosec, and microsec values."""
        tests_ts = [
            (TimeOffset(1, 1000000), "to_millisec", (), 1001),
            (TimeOffset(1, 1000), "to_microsec", (), 1000001),
            (TimeOffset(1, 1), "to_nanosec", (), 1000000001),
            (TimeOffset, 'from_millisec', (1001,), TimeOffset(1, 1000000)),
            (TimeOffset, 'from_microsec', (1000001,), TimeOffset(1, 1000)),
            (TimeOffset, 'from_nanosec', (1000000001,), TimeOffset(1, 1)),
            (TimeOffset(1, 500000), 'to_millisec', (TimeOffset.ROUND_DOWN,), 1000),
            (TimeOffset(1, 500000), 'to_millisec', (TimeOffset.ROUND_NEAREST,), 1001),
            (TimeOffset(1, 499999), 'to_millisec', (TimeOffset.ROUND_NEAREST,), 1000),
            (TimeOffset(1, 500000), 'to_millisec', (TimeOffset.ROUND_UP,), 1001),
            (TimeOffset(1, 500000, -1), 'to_millisec', (TimeOffset.ROUND_DOWN,), -1001),
            (TimeOffset(1, 500000, -1), 'to_millisec', (TimeOffset.ROUND_NEAREST,), -1001),
            (TimeOffset(1, 499999, -1), 'to_millisec', (TimeOffset.ROUND_NEAREST,), -1000),
            (TimeOffset(1, 500000, -1), 'to_millisec', (TimeOffset.ROUND_UP,), -1000)
        ]

        for t in tests_ts:
            with self.subTest(t=t):
                r = getattr(t[0], t[1])(*t[2])
                self.assertEqual(r, t[3],
                                 msg="{!r}.{}{!r} == {!r}, expected {!r}".format(t[0], t[1], t[2], r, t[3]))

    def test_interval_frac(self):
        """This tests that TimeOffsets can be converted to interval fractions."""
        tests_ts = [
            ((50, 1, 1), TimeOffset(0, 20000000)),
            ((50, 1, 2), TimeOffset(0, 10000000))
        ]

        for t in tests_ts:
            with self.subTest(t=t):
                r = TimeOffset.get_interval_fraction(*t[0])
                self.assertEqual(r, t[1],
                                 msg=("TimeOffset.get_interval_fraction{!r} == {!r}, expected {!r}"
                                      .format(t[0], r, t[1])))

        bad_params = [(0, 1, 1),
                      (50, 0, 1),
                      (50, 1, 0)]

        for params in bad_params:
            with self.assertRaises(TsValueError,
                                   msg=("TimeOffset.get_interval_fraction{!r} should have raised TsValueError exception"
                                        .format(params))):
                TimeOffset.get_interval_fraction(*params)

    def test_from_sec_frac(self):
        """This tests that timeoffsets can be instantiated from fractional second values."""
        tests_ts = [
            (("1.000000001",), TimeOffset(1, 1)),
            (("-1.000000001",), TimeOffset(1, 1, sign=-1)),
            (("1.000001POTATO",), TimeOffset(1, 1000)),
            (("1",), TimeOffset(1, 0)),
        ]

        for t in tests_ts:
            with self.subTest(t=t):
                r = TimeOffset.from_sec_frac(*t[0])
                self.assertEqual(r, t[1],
                                 msg="TimeOffset.from_sec_frac{!r} == {!r}, expected {!r}".format(t[0], r, t[1]))

        bad_params = [("0.0.1",), ]

        for params in bad_params:
            with self.assertRaises(TsValueError):
                TimeOffset.from_sec_frac(*params)

    def test_from_sec_nsec(self):
        """This tests that time offsets can be created from second:nanosecond pairs."""
        tests_ts = [
            (("1:1",), TimeOffset(1, 1)),
            (("-1:1",), TimeOffset(1, 1, sign=-1)),
            (("1",), TimeOffset(1, 0)),
        ]

        for t in tests_ts:
            with self.subTest(t=t):
                r = TimeOffset.from_sec_nsec(*t[0])
                self.assertEqual(r, t[1],
                                 msg="TimeOffset.from_sec_nsec{!r} == {!r}, expected {!r}".format(t[0], r, t[1]))

        bad_params = [("0:0:1",), ]

        for params in bad_params:
            with self.assertRaises(TsValueError):
                TimeOffset.from_sec_nsec(*params)

    def test_from_count(self):
        """This tests that time offsets can be created from counts at a specified frequency."""
        tests_ts = [
            ((1, 50, 1), TimeOffset(0, 20000000)),
            ((75, 50, 1), TimeOffset(1, 500000000)),
            ((-75, 50, 1), TimeOffset(1, 500000000, -1))
        ]

        for t in tests_ts:
            r = TimeOffset.from_count(*t[0])
            self.assertEqual(r, t[1],
                             msg="TimeOffset.from_count{!r} == {!r}, expected {!r}".format(t[0], r, t[1]))

        bad_params = [(1, 0, 1),
                      (1, 1, 0)]

        for params in bad_params:
            with self.assertRaises(TsValueError):
                TimeOffset.from_count(*params)

    def test_from_millisec(self):
        """This tests that time offsets can be created from millisecond values."""
        tests_ts = [
            ((1,), TimeOffset(0, 1000000)),
            ((1000,), TimeOffset(1, 0)),
            ((-1,), TimeOffset(0, 1000000, -1))
        ]

        for t in tests_ts:
            r = TimeOffset.from_millisec(*t[0])
            self.assertEqual(r, t[1],
                             msg="TimeOffset.from_millisec{!r} == {!r}, expected {!r}".format(t[0], r, t[1]))

    def test_from_microsec(self):
        """This tests that time offsets can be created from microsecond values."""
        tests_ts = [
            ((1,), TimeOffset(0, 1000)),
            ((1000000,), TimeOffset(1, 0)),
            ((-1,), TimeOffset(0, 1000, -1))
        ]

        for t in tests_ts:
            r = TimeOffset.from_microsec(*t[0])
            self.assertEqual(r, t[1],
                             msg="TimeOffset.from_microsec{!r} == {!r}, expected {!r}".format(t[0], r, t[1]))

    def test_from_nanosec(self):
        """This tests that time offsets can be created from nanosecond values."""
        tests_ts = [
            ((1,), TimeOffset(0, 1)),
            ((1000000000,), TimeOffset(1, 0)),
            ((-1,), TimeOffset(0, 1, -1))
        ]

        for t in tests_ts:
            r = TimeOffset.from_nanosec(*t[0])
            self.assertEqual(r, t[1],
                             msg="TimeOffset.from_nanosec{!r} == {!r}, expected {!r}".format(t[0], r, t[1]))

    def test_set_value(self):
        """This tests that time offsets cannot have their value set."""
        tests_ts = [
            (TimeOffset(0, 0), TimeOffset(0, 1), (0, 1)),
            (TimeOffset(0, 0), TimeOffset(1, 0), (1, 0)),
            (TimeOffset(0, 0), TimeOffset(0, 1, -1), (0, 1, -1))
        ]

        for t in tests_ts:
            ts = deepcopy(t[0])
            with self.assertRaises(AttributeError):
                ts.set_value(*t[2])
            self.assertEqual(ts, t[0])

    def test_to_count(self):
        """This tests that time offsets can be converted to counts at particular frequencies."""
        tests_ts = [
            (TimeOffset(0, 20000000), (50, 1), 1),
            (TimeOffset(1, 500000000), (50, 1), 75),
            (TimeOffset(1, 500000000, -1), (50, 1), -75),
            # below .5 frame
            (TimeOffset(100, 29999999), (50, 1), 100 * 50 + 1),
            # at .5 frame
            (TimeOffset(100, 30000000), (50, 1), 100 * 50 + 2),
            # above .5 frame
            (TimeOffset(100, 30000001), (50, 1), 100 * 50 + 2),
            # below negative .5 frame
            (TimeOffset(100, 9999999), (50, 1), 100 * 50),
            # at negative .5 frame
            (TimeOffset(100, 10000000), (50, 1), 100 * 50 + 1),
            # above negative .5 frame
            (TimeOffset(100, 10000001), (50, 1), 100 * 50 + 1),
            # below .5 frame, round up
            (TimeOffset(100, 29999999), (50, 1, TimeOffset.ROUND_UP), 100 * 50 + 2),
            # at .5 frame, round down
            (TimeOffset(100, 30000000), (50, 1, TimeOffset.ROUND_DOWN), 100 * 50 + 1),
            # above .5 frame, round down
            (TimeOffset(100, 30000001), (50, 1, TimeOffset.ROUND_DOWN), 100 * 50 + 1),
            # below .5 frame, round up
            (TimeOffset(100, 29999999, -1), (50, 1, TimeOffset.ROUND_DOWN), -100 * 50 - 2),
            # at .5 frame, round down
            (TimeOffset(100, 30000000, -1), (50, 1, TimeOffset.ROUND_UP), -100 * 50 - 1),
            # above .5 frame, round down
            (TimeOffset(100, 30000001, -1), (50, 1, TimeOffset.ROUND_UP), -100 * 50 - 1),
        ]

        for t in tests_ts:
            r = t[0].to_count(*t[1])
            self.assertEqual(r, t[2],
                             msg="{!r}.to_count{!r} == {!r}, expected {!r}".format(t[0], t[1], r, t[2]))

        bad_params = [(1, 0),
                      (0, 1)]

        for params in bad_params:
            with self.assertRaises(TsValueError):
                TimeOffset(0, 0).to_count(*params)

    def test_to_microsec(self):
        """This tests that time offsets can be converted to microsecond values."""
        tests_ts = [
            (TimeOffset(0, 1000), (), 1),
            (TimeOffset(1, 1000000), (), 1001000),
            (TimeOffset(1, 1000000, -1), (), -1001000),
            # below .5 us
            (TimeOffset(100, 1499), (), 100 * 1000000 + 1),
            # at .5 us
            (TimeOffset(100, 1500), (), 100 * 1000000 + 2),
            # above .5 us
            (TimeOffset(100, 1501), (), 100 * 1000000 + 2),
            # below .5 us, round up
            (TimeOffset(100, 1499), (TimeOffset.ROUND_UP,), 100 * 1000000 + 2),
            # at .5 us, round up
            (TimeOffset(100, 1500), (TimeOffset.ROUND_UP,), 100 * 1000000 + 2),
            # above .5 us, round up
            (TimeOffset(100, 1501), (TimeOffset.ROUND_UP,), 100 * 1000000 + 2),
            # below .5 us, round down
            (TimeOffset(100, 1499), (TimeOffset.ROUND_DOWN,), 100 * 1000000 + 1),
            # at .5 us, round down
            (TimeOffset(100, 1500), (TimeOffset.ROUND_DOWN,), 100 * 1000000 + 1),
            # above .5 us, round down
            (TimeOffset(100, 1501), (TimeOffset.ROUND_DOWN,), 100 * 1000000 + 1),
            # below .5 us, round down
            (TimeOffset(100, 1499, -1), (TimeOffset.ROUND_DOWN,), -100 * 1000000 - 2),
            # at .5 us, round down
            (TimeOffset(100, 1500, -1), (TimeOffset.ROUND_DOWN,), -100 * 1000000 - 2),
            # above .5 us, round down
            (TimeOffset(100, 1501, -1), (TimeOffset.ROUND_DOWN,), -100 * 1000000 - 2),
            # below .5 us, round up
            (TimeOffset(100, 1499, -1), (TimeOffset.ROUND_UP,), -100 * 1000000 - 1),
            # at .5 us, round up
            (TimeOffset(100, 1500, -1), (TimeOffset.ROUND_UP,), -100 * 1000000 - 1),
            # above .5 us, round up
            (TimeOffset(100, 1501, -1), (TimeOffset.ROUND_UP,), -100 * 1000000 - 1),
        ]

        for t in tests_ts:
            r = t[0].to_microsec(*t[1])
            self.assertEqual(r, t[2],
                             msg="{!r}.to_microsec{!r} == {!r}, expected {!r}".format(t[0], t[1], r, t[2]))

    def test_abs(self):
        """This tests that negative time offsets can be converted to positive ones using abs."""
        tests_ts = [
            (TimeOffset(10, 1), TimeOffset(10, 1)),
            (TimeOffset(10, 1, -1), TimeOffset(10, 1))
        ]

        for t in tests_ts:
            r = abs(t[0])
            self.assertEqual(r, t[1],
                             msg="abs({!r}) == {!r}, expected {!r}".format(t[0], r, t[1]))

    def test_average(self):
        """This tests that time offsets can be averaged."""
        toff1 = TimeOffset(11, 976)
        toff2 = TimeOffset(21, 51)
        toff_avg = (toff1 * 49 + toff2) // 50
        avg = int((toff1.to_nanosec() * 49 + toff2.to_nanosec()) // 50)
        self.assertEqual(avg, toff_avg.to_nanosec())

    def test_cast(self):
        """This tests that addition and subtraction of TimeOffsets and integers or floats works as expected."""
        tests_ts = [
            (TimeOffset(10, 1), '+',  1, TimeOffset(11, 1)),
            (TimeOffset(10, 1), '-', 1, TimeOffset(9, 1)),
            (TimeOffset(10, 1), '+', 1.5, TimeOffset(11, 500000001)),
            (TimeOffset(10, 1), '-', 1.5, TimeOffset(8, 500000001)),
        ]

        for t in tests_ts:
            if t[1] == '+':
                r = t[0] + t[2]
            else:
                r = t[0] - t[2]
            self.assertEqual(r, t[3],
                             msg="{!r} {} {!r} == {!r}, expected {}".format(t[0], t[1], t[2], r, t[3]))

        self.assertEqual(TimeOffset(8, 500000000), 8.5)
        self.assertGreater(TimeOffset(8, 500000000), 8)
        self.assertLess(TimeOffset(8, 500000000), 8.6)
        self.assertNotEqual(TimeOffset(8, 500000000), 8.6)

    def test_eq(self):
        """Test some equality operations"""
        self.assertEqual(TimeOffset(10, 1), TimeOffset(10, 1))
        self.assertNotEqual(TimeOffset(10, 1), TimeOffset(-10, 1))
        self.assertNotEqual(TimeOffset(0, 1), TimeOffset(0, -1))
        self.assertNotEqual(TimeOffset(0, 1), None)
        self.assertEqual(TimeOffset(0, 500000000), 0.5)
        self.assertNotEqual(TimeOffset(0, 1), "0:1")

    def test_str(self):
        """This tests that the str function turns time offsets into second:nanosecond pairs."""
        tests_ts = [
            (str(TimeOffset(10, 1)), "10:1"),
            (str(TimeOffset(10, 1, -1)), "-10:1"),
        ]

        for t in tests_ts:
            self.assertEqual(t[0], t[1])



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from setuptools import setup
import os

# Basic metadata
name = 'mediatimestamp'
version = '2.2.1'
description = 'A timestamp library for high precision nanosecond timestamps'
url = 'https://github.com/bbc/rd-apmm-python-lib-mediatimestamp'
author = 'James P. Weaver'
author_email = 'james.barrett@bbc.co.uk'
license = 'Apache 2'
long_description = description


def is_package(path):
    return (
        os.path.isdir(path) and
        os.path.isfile(os.path.join(path, '__init__.py'))
        )


def find_packages(path, base=""):
    """ Find all packages in path """
    packages = {}
    for item in os.listdir(path):
        dir = os.path.join(path, item)
        if is_package(dir):
            if base:
                module_name = "%(base)s.%(item)s" % vars()
            else:
                module_name = item
            packages[module_name] = dir
            packages.update(find_packages(dir, module_name))
    return packages


packages = find_packages(".")
package_names = packages.keys()

# This is where you list packages which are required
packages_required = [
    "python-dateutil>=2.1,<2.8.1"
]

# This is where you list locations for packages not
# available from pip. Each entry must be of the form:
#  <url>#egg=<pkgname>=<version>
# eg. https://github.com/bbc/rd-apmm-python-lib-nmos-common#egg=nmoscommon=0.1.0
deps_required = []

setup(name=name,
      python_requires='>=3.6.0',
      version=version,
      description=description,
      url=url,
      author=author,
      author_email=author_email,
      license=license,
      packages=package_names,
      package_dir=packages,
      package_data={name: ['py.typed'] for name in package_names},
      install_requires=packages_required,
      scripts=[],
      data_files=[],
      long_description=long_description)
